from pymonet.monad_try import Try


def divide(dividend, divisor):
    return dividend / divisor


def test_try_should_call_success_callback_with_result_of_function_when_exception_was_not_thrown():

    def success_callback(value):
        assert value == 21

    def fail_callback(_):
        assert True is False

    (Try.of(divide, 42, 2)
        .on_success(success_callback)
        .on_fail(fail_callback))


def test_try_should_call_fail_callback_with_result_of_function_when_exception_was_thrown():

    def success_callback(_):
        assert True is False

    def fail_callback(error):
        assert isinstance(error, ZeroDivisionError)
        assert str(error) == 'division by zero'

    (Try.of(divide, 42, 0)
        .on_success(success_callback)
        .on_fail(fail_callback))


def test_try_eq_should_compare_value_and_result_of_try():
    assert Try.of(divide, 42, 2) == Try.of(divide, 42, 2)
    assert Try.of(divide, 43, 2) != Try.of(divide, 42, 2)
    assert Try.of(divide, 43, 1) != Try.of(divide, 42, 0)


def test_try_should_appied_map_when_exception_was_thrown():

    def success_callback(value):
        assert value == 22

    def fail_callback(_):
        assert True is False

    (Try.of(divide, 42, 2)
        .map(lambda value: value + 1)
        .on_success(success_callback)
        .on_fail(fail_callback))


def test_try_should_appied_fold_when_exception_not_was_thrown():

    def fail_callback(_):
        assert True is False

    (Try.of(divide, 42, 2)
        .fold(lambda value: Try.of(divide, 42, 3))
        .on_success(lambda value: value == 7)
        .on_fail(fail_callback))


def test_try_should_not_appied_fold_when_exception_was_thrown():

    def success_callback(_):
        assert True is False

    def fail_callback(error):
        assert isinstance(error, ZeroDivisionError)
        assert str(error) == 'division by zero'

    (Try.of(divide, 42, 0)
        .fold(lambda value: Try.of(divide, 42, 3))
        .on_success(success_callback)
        .on_fail(fail_callback))


def test_when_fold_is_rejected_monad_also_should_be_rejected():

    def success_callback(_):
        assert True is False

    def fail_callback(error):
        assert isinstance(error, ZeroDivisionError)
        assert str(error) == 'float division by zero'

    (Try.of(divide, 42, 2)
        .fold(lambda value: Try.of(divide, value, 0))
        .on_success(success_callback)
        .on_fail(fail_callback))


def test_try_should_not_applied_map_when_exception_thrown():

    def success_callback(_):
        assert True is False

    def fail_callback(error):
        assert isinstance(error, ZeroDivisionError)
        assert str(error) == 'division by zero'

    def mapper(_):
        assert True is False

    (Try.of(divide, 42, 0)
        .map(mapper)
        .on_success(success_callback)
        .on_fail(fail_callback))


def test_get_or_default_method_should_return_value_when_exception_was_not_thrown():
    assert Try.of(divide, 42, 2).get_or_else('Holy Grail') == 21


def test_get_or_default_method_should_return_default_value_when_exception_was_thrown():
    assert Try.of(divide, 42, 0).get_or_else('Holy Grail') == 'Holy Grail'


def test_get_method_should_return_value_with_or_without_exception_thrown():
    assert Try.of(divide, 42, 2).get() == 21

    failed_value = Try.of(divide, 42, 0).get()
    assert isinstance(failed_value, ZeroDivisionError)
    assert str(failed_value) == 'division by zero'


def test_filer_should_converts_to_fail_when_predicate_returns_false():
    filtered = Try.of(divide, 42, 2).filter(lambda value: value % 2 == 0.0)
    assert not filtered.is_success

    not_filtered = Try.of(divide, 42, 2).filter(lambda value: value % 3 == 0.0)
    assert not_filtered.is_success



from pymonet.task import Task


class TaskSpy:

    def resolved(self, value):
        return value

    def rejected(self, value):
        return value

    def mapper(self, arg):
        return arg + 1

    def side_effect(self, arg):  # remove?
        return arg

    def folder(self, arg):
        return Task(lambda reject, resolve: resolve(arg + 1))

    def fork_rejected(self, reject, _):
        reject(0)

    def fork_resolved(self, _, resolve):
        resolve(42)


def test_task_resolved_fork_should_be_called_only_during_calling_fork(mocker):
    task_spy = TaskSpy()
    mocker.spy(task_spy, 'resolved')
    mocker.spy(task_spy, 'rejected')
    mocker.spy(task_spy, 'fork_rejected')
    mocker.spy(task_spy, 'fork_resolved')

    task = Task(task_spy.fork_resolved)
    assert task_spy.fork_resolved.call_count == 0

    task.fork(task_spy.rejected, task_spy.resolved)
    assert task_spy.fork_resolved.call_count == 1
    assert task_spy.fork_rejected.call_count == 0
    assert task_spy.resolved.call_count == 1
    assert task_spy.rejected.call_count == 0


def test_task_rejected_fork_should_be_called_only_during_calling_fork(mocker):
    task_spy = TaskSpy()
    mocker.spy(task_spy, 'resolved')
    mocker.spy(task_spy, 'rejected')
    mocker.spy(task_spy, 'fork_rejected')
    mocker.spy(task_spy, 'fork_resolved')

    task = Task(task_spy.fork_rejected)
    assert task_spy.fork_resolved.call_count == 0

    task.fork(task_spy.rejected, task_spy.resolved)
    assert task_spy.fork_resolved.call_count == 0
    assert task_spy.fork_rejected.call_count == 1
    assert task_spy.fork_resolved.call_count == 0
    assert task_spy.fork_rejected.call_count == 1


def test_task_resolved_fork_should_return_resolved_value():
    def rejected_spy(_):
        assert False

    def resolved_spy(value):
        assert value == 42

    task_spy = TaskSpy()
    task = Task(task_spy.fork_resolved)
    task.fork(rejected_spy, resolved_spy)


def test_task_rejected_fork_should_return_resolved_value():
    def rejected_spy(value):
        assert value == 0

    def resolved_spy(_):
        assert False

    task_spy = TaskSpy()
    task = Task(task_spy.fork_rejected)
    task.fork(rejected_spy, resolved_spy)


def test_task_maper_should_be_called_during_calling_fork(mocker):
    task_spy = TaskSpy()
    mocker.spy(task_spy, 'mapper')

    task = Task(task_spy.fork_resolved).map(task_spy.mapper)
    assert task_spy.mapper.call_count == 0

    task.fork(task_spy.rejected, task_spy.resolved)
    assert task_spy.mapper.call_count == 1


def test_task_rejected_fork_should_not_applied_map_on_his_result():
    def rejected_spy(value):
        assert value == 0

    def resolved_spy(_):
        assert False

    task_spy = TaskSpy()
    task = Task(task_spy.fork_rejected).map(task_spy.mapper)
    task.fork(rejected_spy, resolved_spy)


def test_task_resolved_fork_should_applied_map_on_his_result():
    def rejected_spy(_):
        assert False

    def resolved_spy(value):
        assert value == 42 + 1

    task_spy = TaskSpy()
    task = Task(task_spy.fork_resolved).map(task_spy.mapper)
    task.fork(rejected_spy, resolved_spy)


def test_task_rejected_fork_should_not_applied_fold_on_his_result():
    def rejected_spy(value):
        assert value == 0

    def resolved_spy(_):
        assert False

    task_spy = TaskSpy()
    task = Task(task_spy.fork_rejected).fold(task_spy.folder)
    task.fork(rejected_spy, resolved_spy)


def test_task_resolved_fork_should_applied_fold_on_his_result():
    def rejected_spy(_):
        assert False

    def resolved_spy(value):
        assert value == 42 + 1

    task_spy = TaskSpy()
    task = Task(task_spy.fork_resolved).fold(task_spy.folder)
    task.fork(rejected_spy, resolved_spy)



class Either:
    """
    The Either type represents values with two possibilities: B value of type Either<A, B> is either Left<A> or Right<B>
    But not both in the same time.
    """
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return isinstance(other, Either) and\
            self.value == other.value and\
            self.is_right() == other.is_right()

    def case(self, error, success):
        """
        takes 2 functions call only one of then with either value and return her result
        :params error: function to call when Either is Left
        :type (A) -> B
        :params success: function to call when Either is Right
        :type (A) -> B
        :return: B
        """
        if self.is_right():
            return success(self.value)
        return error(self.value)


class Left(Either):

    def map(self, _):
        """
        takes mapper function and return new instance of Left with the same value
        :return: Left<A>
        """
        return Left(self.value)

    def fold(self, _):
        """
        takes mapper function and return value of Left
        :return: A
        """
        return self.value

    def is_left(self):
        """
        :return: Boolean
        """
        return True

    def is_right(self):
        """
        :return: Boolean
        """
        return False


class Right(Either):

    def map(self, mapper):
        """
        takes mapper function and return new instance of Right with mapped value
        :param mapper: function to apply on Right value
        :type (A) -> B
        :return: Right<B>
        """
        return Right(mapper(self.value))

    def fold(self, mapper):
        """
        takes mapper function and returns result of them called with Right value
        :param mapper: function to apply on Right value
        :type (A) -> Either<B>
        :return: Either<B>
        """
        return mapper(self.value)

    def is_right(self):
        """
        :return: Boolean
        """
        return True

    def is_left(self):
        """
        :return: Boolean
        """
        return False



from functools import reduce


def identity(value):
    return value


def increase(value):
    return value + 1


def curried_map(mapper):
    return lambda collection: [mapper(item) for item in collection]


def curried_filter(filterer):
    return lambda collection: [item for item in collection if filterer(item)]


def compose(value, *functions):
    """
    Performs right-to-left function composition.
    :param value - argument of first applied function
    :type Any
    :param functions - list of functions to applied from right-to-left
    :return result of all functions
    :typa Any
    """
    return reduce(
        lambda current_value, function: function(current_value),
        functions[::-1],
        value
    )


def pipe(value, *functions):
    """
    Performs ledt-to-right function composition.
    :param value - argument of first applied function
    :type Any
    :param functions - list of functions to applied from ledt-to-right
    :return result of all functions
    :typa Any
    """
    return reduce(
        lambda current_value, function: function(current_value),
        functions,
        value
    )



from pymonet.box import Box


def test_eq_should_compare_only_box_value():
    assert Box(42) == Box(42)
    assert Box(43) != Box(42)
    assert Box([]) == Box([])
    assert Box({}) == Box({})
    assert Box(None) == Box(None)


def test_map_should_return_box_with_mapped_value():
    box = Box(42)
    assert box.map(lambda value: value + 1) == Box(43)


def test_map_should_return_new_instance_of_box():
    box = Box(42)
    mapped_box = box.map(lambda value: value + 1)
    assert box is not mapped_box


def test_fold_should_return_result_of_fold_function_called_with_box_value():
    box = Box(42)
    assert box.fold(lambda value: value + 1) == 43



class Try:

    def __init__(self, value, is_success):
        self.value = value
        self.is_success = is_success

    def __eq__(self, other):
        return isinstance(other, type(self))\
            and self.value == other.value\
            and self.is_success == other.is_success

    @classmethod
    def of(cls, fn, *args):
        try:
            return cls(fn(*args), True)
        except Exception as e:
            return cls(e, False)

    def map(self, mapper):
        if self.is_success:
            return Try(mapper(self.value), True)
        return Try(self.value, False)

    def fold(self, mapper):
        if self.is_success:
            return mapper(self.value)
        return self

    def on_success(self, success_callback):
        if self.is_success:
            success_callback(self.value)
        return self

    def on_fail(self, fail_callback):
        if not self.is_success:
            fail_callback(self.value)
        return self

    def filter(self, filterer):
        if self.is_success and filterer(self.value):
            return Try(self.value, True)
        return Try(self.value, False)

    def get(self):
        return self.value

    def get_or_else(self, default_value):
        if self.is_success:
            return self.value
        return default_value



from pymonet.maybe import Maybe
from pymonet.utils import increase


def wrong_mapper(_):
    assert True is False


def test_maybe_eq_operator_shuld_compare_values():

    assert Maybe.just(42) == Maybe.just(42)
    assert Maybe.just(None) == Maybe.just(None)

    assert Maybe.just(42) != Maybe.nothing()


def test_maybe_map_operator_should_be_applied_only_on_just_value():
    assert Maybe.just(42).map(increase) == Maybe.just(43)
    assert Maybe.nothing().map(increase) == Maybe.nothing()


def test_maybe_map_should_not_call_mapper_when_monad_has_nothing():
    Maybe.nothing().map(wrong_mapper)


def test_maybe_fold_should_retrun_result_of_mapper_called_with_maybe_value():
    assert Maybe.just(42).fold(increase) == 43


def test_maybe_fold_should_not_call_mapper_when_monad_has_nothing():
    Maybe.nothing().fold(wrong_mapper)


def test_maybe_get_or_else_method_should_return_maybe_value_when_monad_is_not_empy():
    assert Maybe.just(42).get_or_else(0) == 42


def test_maybe_get_or_else_method_should_return_argument_when_monad_is_emtpy():
    assert Maybe.nothing().get_or_else(0) == 0


def test_maybe_is_nothing_should_return_proper_boolean():
    assert Maybe.just(42).is_nothing is False
    assert Maybe.nothing().is_nothing is True


def test_maybe_if_filter_returns_false_method_should_return_empty_maybe():
    assert Maybe.just(41).filter(lambda value: value % 2 == 0) == Maybe.nothing()


def test_maybe_if_filter_returns_true_method_should_return_self():
    assert Maybe.just(42).filter(lambda value: value % 2 == 0) == Maybe.just(42)



class Box:
    """
    Data type for storage any type of data
    """

    def __init__(self, value):
        """
        :param value: value to store in Box
        :type any
        """
        self.value = value

    def __eq__(self, other):
        return self.value == other.value

    def map(self, fn):
        """
        takes function (a) -> b and applied this function on current box value and returns new box with mapped value
        :param fn: mapper function
        :type (a) -> b
        :return: new box with mapped value
        :type Box<b>
        """
        return Box(fn(self.value))

    def fold(self, fn):
        """
        takes function (a) -> b and applied this function on current box value and returns mapped value
        :param fn: mapper function
        :type (a) -> b
        :return: new box with mapped value
        :type b
        """
        return fn(self.value)



from pymonet.applicative import Applicative


class ApplicativeSpy:

    def fn(self):
        return 42

    def side_effect(self, input):
        return input


def test_applicative(mocker):
    """
    input function and all map functions should be called during called fold method
    """

    applicative_spy = ApplicativeSpy()
    mocker.spy(applicative_spy, 'fn')
    mocker.spy(applicative_spy, 'side_effect')

    applicative = Applicative(applicative_spy.fn)
    assert applicative_spy.fn.call_count == 0

    applicative = applicative.map(applicative_spy.side_effect)
    assert applicative_spy.fn.call_count == 0
    assert applicative_spy.side_effect.call_count == 0

    applicative = applicative.map(lambda number: number + 1)
    assert applicative_spy.fn.call_count == 0
    assert applicative_spy.side_effect.call_count == 0

    assert applicative.fold(lambda number: number + 1) == 44
    assert applicative_spy.fn.call_count == 1
    assert applicative_spy.side_effect.call_count == 1



class Task:

    def __init__(self, fork):
        """
        :param fork: function to call during fork
         :type: (reject, resolve) -> any
        """
        self.fork = fork

    @classmethod
    def of(cls, value):
        """
        :param value
        :type: any
        instant rejected Task
        :return: Task<_, resolve>
        """
        return lambda _, resolve: resolve(value)

    @classmethod
    def reject(cls, value):
        """
        :param value
        :type: any
        instant rejected Task
        :return: Task<reject, _>
        """
        return lambda reject, _: reject(value)

    def map(self, fn):
        """
        :param fn: mapper function
        :type fn: value -> mapped_value
        :return: Task<reject -> mapped_value>
        """
        def result(reject, resolve):
            return self.fork(
                lambda arg: reject(arg),
                lambda arg: resolve(fn(arg))
            )

        return Task(result)

    def fold(self, fn):
        """
        also know as flatmap
        :param fn: mapper function
        :type fn: value -> Task<reject, mapped_value>
        :return: Task<reject, mapped_value>
        """
        def result(reject, resolve):
            return self.fork(
                lambda arg: reject(arg),
                lambda arg: fn(arg).fork(reject, resolve)
            )

        return Task(result)



from pymonet.semigroups import Sum, All, First, Map


ingredient1 = Map({'score': Sum(1), 'won': All(True), 'captain': First('captain america')})
ingredient2 = Map({'score': Sum(2), 'won': All(True), 'captain': First('iron man')})
ingredient3 = Map({'score': Sum(3), 'won': All(False), 'captain': First('Batman')})


def test_sum():
    assert Sum(1).concat(Sum(2)) == Sum(3)
    assert Sum(1).concat(
        Sum(2).concat(Sum(3))
    ) == Sum(6)
    assert Sum(1).concat(Sum(2)).concat(Sum(3)) == Sum(6)


def test_all():
    assert All(True).concat(All(True)) == All(True)
    assert All(True).concat(All(False)) == All(False)

    assert All(True).concat(
        All(True).concat(All(True))
    ) == All(True)
    assert All(True).concat(All(True)).concat(All(True)) == All(True)

    assert All(True).concat(
        All(False).concat(All(True))
    ) == All(False)
    assert All(True).concat(All(True)).concat(All(False)) == All(False)


def test_first():
    assert First('first').concat(First('second')) == First('first')
    assert First('first').concat(First('second')).concat(First('third')) == First('first')
    assert First('first').concat(
        First('second').concat(First('third'))
    ) == First('first')


def test_map3():
    assert ingredient1.concat(ingredient2) == Map(
        {'score': Sum(3), 'won': All(True), 'captain': First('captain america')}
    )


def test_map1():
    assert ingredient1.concat(ingredient2).concat(ingredient3) == Map(
        {'score': Sum(6), 'won': All(False), 'captain': First('captain america')})


def test_map2():
    assert ingredient1.concat(ingredient2.concat(ingredient3)) == Map(
        {'score': Sum(6), 'won': All(False), 'captain': First('captain america')})



class Semigroup:

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value

    def fold(self, fn):
        return fn(self.value)


class Sum(Semigroup):

    def concat(self, semigroup):
        return Sum(self.value + semigroup.value)


class All(Semigroup):

    def concat(self, semigroup):
        return All(self.value and semigroup.value)


class First(Semigroup):

    def concat(self, semigroup):
        return First(self.value)


class Map(Semigroup):

    def concat(self, semigroup):
        return Map(
            {key: value.concat(semigroup.value[key]) for key, value in self.value.items()}
        )



from pymonet.utils import \
    identity,\
    increase,\
    compose,\
    pipe,\
    curried_map as map,\
    curried_filter as filter


def test_identity_should_return_first_argument():
    assert identity(42) is 42
    assert identity('string') is 'string'
    assert identity([1, 2, 3, 4, 5]) == [1, 2, 3, 4, 5]
    assert identity(None) is None


def test_compose_should_applied_function_on_value_and_return_it_result():
    assert compose(42, increase) == 43
    assert compose(42, increase, increase) == 44
    assert compose(42, increase, increase, increase) == 45


def test_compose_should_appield_functions_from_last_to_first():
    assert compose(42, increase, lambda value: value * 2) == 85


def test_compose_with_collections():
    def is_odd(value): return value % 2 == 0

    assert compose(
        list(range(10)),
        map(increase),
        filter(is_odd)
    ) == [1, 3, 5, 7, 9]

    assert compose(
        list(range(10)),
        filter(is_odd),
        map(increase)
    ) == [2, 4, 6, 8, 10]


def test_pipe_should_appield_functions_from_first_to_last():
    assert pipe(42, increase, lambda value: value * 2) == 86



class Maybe():

    def __init__(self, value, is_nothing):
        self.is_nothing = is_nothing
        if not is_nothing:
            self.value = value

    def __eq__(self, other):
        return isinstance(other, Maybe) and \
            self.is_nothing == other.is_nothing and \
            (self.is_nothing or self.value == other.value)

    @classmethod
    def just(cls, value):
        """
        creates not empty maybe
        :params value: value to store in Maybe
        :type Any
        :return Maybe<Any>
        """
        return Maybe(value, False)

    @classmethod
    def nothing(cls):
        """
        creates empty maybe
        :return Maybe<None>
        """
        return Maybe(None, True)

    def map(self, mapper):
        """
        if Maybe is empty return new empty Maybe, in other case
        takes mapper function and returns new instance of Maybe
        with result of mapper
        :params mapper: function to call with Maybe value
        :type (A) -> B
        :return Maybe<B | None>
        """
        if self.is_nothing:
            return Maybe.nothing()
        return Maybe.just(
            mapper(self.value)
        )

    def fold(self, mapper):
        """
        if Maybe is empty return new empty Maybe, in other case
        takes mapper function and returns result of mapper
        :params mapper: function to call with Maybe.value
        :type (A) -> Maybe<B>
        :return Maybe<B | None>
        """
        if self.is_nothing:
            return Maybe.nothing()
        return mapper(self.value)

    def filter(self, filterer):
        """
        if Maybe is empty or filterer returns False return default_value, in other case
        return new instance of Maybe with the same value
        :params filterer:
        :type (A) -> Boolean
        :returns Maybe<A> | Maybe<None>
        """
        if self.is_nothing or not filterer(self.value):
            return Maybe.nothing()
        return Maybe.just(self.value)

    def get_or_else(self, default_value):
        """
        if Maybe is empty return default_value, in other case
        returns Maybe.value
        :params default_value: value to return if Maybe is empty
        :type Any
        :returns A | Any
        """
        if self.is_nothing:
            return default_value
        return self.value



from setuptools import setup


setup(
  name='pymonet',
  packages=['pymonet'],
  version='0.5.0',
  description='High abstract python library for functional programming. Contains algebraic data structures known from Haskell or Scala.',
  author='Przemyslaw Jan Pietrzak',
  author_email='przemyslaw.jan.pietrzak@gmail.com',
  url='https://github.com/przemyslawjanpietrzak/pyMonet',
  download_url='https://github.com/przemyslawjanpietrzak/pyMonet/archive/0.5.0.tar.gz',
  keywords=['functional-programming', 'monads', 'functors', 'semigroups'],
  classifiers=[],
)



class Applicative:
    """
     Data type for storage any type of function.
     This function (and all his mappers) will be called only during calling fold method
     """

    def __init__(self, constructor_fn):
        """
        :param constructor_fn: function to call during fold method call
        """
        self.constructor_fn = constructor_fn

    def map(self, fn):
        """
        takes function (A) -> A and returns new LazyBox with mapped argument of Applicative constructor function.
        Both mapper end constructor will be called only during calling fold method
        :param fn: mapper function
        :type   (constructor_fn) -> B
        :return: Applicative<() -> fn(constructor_fn)>
        """
        return Applicative(lambda: fn(self.constructor_fn()))

    def fold(self, fn):
        """
        takes function and call constructor function passing returned value to fn function.
        It's only way to call function store in Applicative
        :param fn: (constructor_fn) -> B
        :return: B
        """
        return fn(self.constructor_fn())



from pymonet.either import Left, Right
from pymonet.utils import increase


class EitherSpy:

    def error_handler(value, *args):
        pass

    def success_handler(value, *args):
        pass


def test_either_eq_operator_should_compare_values():

    assert Right(42) == Right(42)
    assert Right(42) != Right(43)

    assert Left(42) == Left(42)
    assert Left(42) != Left(43)

    assert Right(42) != Left(42)


def test_mapper_should_be_applied_only_on_current_value():

    assert Left(42).map(increase) == Left(42)
    assert Right(42).map(increase) == Right(43)


def test_is_right_should_return_suitable_value():

    assert Right(42).is_right()
    assert not Left(42).is_right()


def test_is_left_should_return_suitable_value():
    assert Left(42).is_left()
    assert not Right(42).is_left()


def test_fold_should_be_applied_only_on_current_value_and_return_value():

    assert Left(42).fold(increase) == 42
    assert Right(42).fold(increase) == 43


def test_case_method_should_call_proper_handler(mocker):
    either_spy = EitherSpy()
    mocker.spy(either_spy, 'error_handler')
    mocker.spy(either_spy, 'success_handler')

    Left(42).case(
        success=either_spy.success_handler,
        error=either_spy.error_handler
    )

    assert either_spy.error_handler.call_count == 1
    assert either_spy.success_handler.call_count == 0

    Right(42).case(
        success=either_spy.success_handler,
        error=either_spy.error_handler
    )

    assert either_spy.error_handler.call_count == 1
    assert either_spy.success_handler.call_count == 1
